/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

/* Not strictly needed on non-Solaris, but strcasecmp is POSIX.1-2001 */
#define _XOPEN_SOURCE   500

#include <xacml_client.h>
#include <xacml_io_ssl.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <stdint.h>

#include <errno.h>
#include <grp.h>

#include <errno.h>

/* B: prototypes */

#include "pep_obligation_handlers.h"
#include "xacml_attr_obl_profile.h"

#include <lcmaps/lcmaps_log.h>
#include <lcmaps/lcmaps_modules.h>


/************************************************************************
 * global variables
 ************************************************************************/

/* set/get using pep_{get,set}_use_authorization_only */
static int use_authorization_only = 0;

/* globals to keep track of which OH has fired */
static int oh_uidgid_fired        = 0;
static int oh_secondarygids_fired = 0;
static int oh_username_fired      = 0;
static int oh_account_fired       = 0;

/* variables to contain the internal list of credentials */
static uid_t *uid_list=NULL;
static gid_t *pgid_list=NULL;
static gid_t *sgid_list=NULL;
static size_t cnt_uid=0;
static size_t max_cnt_uid=0;
static size_t cnt_pgid=0;
static size_t max_cnt_pgid=0;
static size_t cnt_sgid=0;
static size_t max_cnt_sgid=0;


/************************************************************************
 * private prototypes
 ************************************************************************/

static void addHandler(xacml_request_t request,
		       xacml_obligation_handler_t handler,
		       const char *obligationID);

/* Setter functions for globals, to keep track of which OHs have run */
static void setOHUIDGIDFired(int);
static void setOHSecondaryGIDsFired(int);
static void setOHUsernameFired(int);
static void setOHAccountFired(int);

/* Add functions to add credentials to the internal lists. */
static int addUid(uid_t uid);
static int addPGid(gid_t pgid);
static int addSGid(gid_t sgid[], size_t count);


/************************************************************************
 * Helper Functions
 ************************************************************************/

/* Setter for use_authorization_only */
void pep_set_use_authorization_only(int value)	{
    use_authorization_only=value;
}

/* Getter for use_authorization_only */
int pep_get_use_authorization_only(void)	{
    return use_authorization_only;
}

/*************************** B: Obligation Handlers *********************************/
int default_handler(
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    const char * logstr = "default_handler";
    int i;

    lcmaps_log(LOG_WARNING, "%s: Unhandled obligation in result: %s\n",
	    logstr, obligation_id);
    for (i = 0; attribute_ids[i] != NULL; i++)
        lcmaps_log (LOG_DEBUG, "%s:    attribute: %s [%s] = %s\n",
		logstr, attribute_ids[i], datatypes[i], values[i]);

    return 1;
}




int UIDGID_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    const char * logstr = "UIDGID_handler";
    int rc = 0;
    int i  = 0;
    uid_t uid;
    gid_t gid;
    long long_value;
    int uid_attribs=0, gid_attribs=0;
    char *endptr = NULL;
    xacml_decision_t decision=XACML_DECISION_Indeterminate;

    /* Get the decision from the response */
    if (xacml_response_get_xacml_decision(response, &decision)!=XACML_RESULT_SUCCESS)	{
	lcmaps_log(LOG_ERR, "%s: Error: Cannot get decision from response\n",
		logstr);
	return 1;
    }

    /* Check that the decision is Permit or Deny and matches the fullfill_on for
     * the given obligation */
    if ( ! ( (decision==XACML_DECISION_Permit &&
	      fulfill_on==XACML_EFFECT_Permit) ||
	     (decision==XACML_DECISION_Deny &&
	      fulfill_on==XACML_EFFECT_Deny) ) )
    {
	lcmaps_log(LOG_WARNING,
		"%s: Cannot fulfill obligation, "
		"fulfill on \"%s\" does not match decision \"%s\".\n",
		logstr, fulfill_on_str(fulfill_on), decision_str(decision));
	return 1;
    }
    
    lcmaps_log(LOG_DEBUG, "%s: Got obligation %s\n", logstr, obligation_id);
    setOHUIDGIDFired(1);

    for (i = 0; attribute_ids[i] != NULL && rc == 0; i++)
    {
	/* Check validity of value */
	if (values[i]==0 || values[i][0]=='\0') {
	    lcmaps_log(LOG_ERR,
		    "%s: Error: Attribute %s has empty value.\n",
		    logstr, attribute_ids[i]);
	    return 1;
	}

        lcmaps_log(LOG_DEBUG, "%s: %s [%s] = %s\n", logstr, attribute_ids[i],
		datatypes[i] ? datatypes[i] : "NULL", values[i]);

        /* posix-uid attribute */
        if (strcasecmp(attribute_ids[i],
		       XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID) == 0)
        {
	    if (uid_attribs!=0)	{
		lcmaps_log(LOG_ERR,
		    "%s: error: Found multiple %s attributes. "
		    "This is a protocol error.\n",
		    logstr, attribute_ids[i]);
		return 1;
	    }
	    /* Set flag */
	    uid_attribs=1;
	    errno=0;
	    long_value=strtol(values[i], &endptr, 10);
	    if (endptr[0]!='\0' || errno!=0 || long_value<0)   {
		lcmaps_log(LOG_ERR,"%s: Cannot convert %s to a uid.\n",
			logstr, values[i]);
		return 1;
	    }
            uid = (uid_t)long_value;
            lcmaps_log(LOG_DEBUG, "%s: Adding UID :  %d\n", logstr, uid);
            if (!use_authorization_only)
                rc = addUid(uid);
        }

        /* posix-gid attribute */
        else if (strcasecmp(attribute_ids[i],
			    XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID) == 0)
        {
	    if (gid_attribs!=0)	{
		lcmaps_log(LOG_ERR,
		    "%s: error: Found multiple %s attributes. "
		    "This is a protocol error.\n",
		    logstr, attribute_ids[i] );
		return 1;
	    }
	    /* Set flag */
	    gid_attribs=1;
	    errno=0;
	    long_value=strtol(values[i], &endptr, 10);
	    if (endptr[0]!='\0' || errno!=0 || long_value<0)   {
		lcmaps_log(LOG_ERR,"%s: Cannot convert %s to a gid.\n",
			logstr, values[i]);
		return 1;
	    }
            gid = (gid_t)long_value;
            lcmaps_log(LOG_DEBUG, "%s: Adding GID :  %d\n", logstr, gid);
            if (!use_authorization_only)
                rc = addPGid(gid);
        }
        else
        {
            lcmaps_log(LOG_ERR,
		    "%s: Error: Unexpected attribute: %s [%s] = %s\n",
		    logstr, attribute_ids[i],
		    datatypes[i] ? datatypes[i] : "NULL", values[i]);
            return 1;
        }
    }

    return rc;
}


int SecondaryGIDs_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    const char * logstr = "SecondaryGIDs_handler";
    int rc = 0;
    int i  = 0;
    gid_t gid;
    long long_value;
    char *endptr=NULL;
    xacml_decision_t decision=XACML_DECISION_Indeterminate;

    /* Get the decision from the response */
    if (xacml_response_get_xacml_decision(response, &decision)!=XACML_RESULT_SUCCESS)	{
	lcmaps_log(LOG_ERR, "%s: Error: Cannot get decision from response\n",
		logstr);
	return 1;
    }

    /* Check that the decision is Permit or Deny and matches the fullfill_on for
     * the given obligation */
    if ( ! ( (decision==XACML_DECISION_Permit &&
	      fulfill_on==XACML_EFFECT_Permit) ||
	     (decision==XACML_DECISION_Deny &&
	      fulfill_on==XACML_EFFECT_Deny) ) )
    {
	lcmaps_log(LOG_WARNING,
		"%s: Cannot fulfill obligation, "
		"fulfill on \"%s\" does not match decision \"%s\".\n",
		logstr, fulfill_on_str(fulfill_on), decision_str(decision));
	return 1;
    }
    
    lcmaps_log(LOG_DEBUG, "%s: Got obligation %s\n", logstr, obligation_id);
    setOHSecondaryGIDsFired(1);

    for (i = 0; attribute_ids[i] != NULL && rc == 0; i++)
    {
	/* Check validity of value */
	if (values[i]==0 || values[i][0]=='\0') {
	    lcmaps_log(LOG_ERR,
		    "%s: Error: Attribute %s has empty value.\n",
		    logstr, attribute_ids[i]);
	    return 1;
	}

        lcmaps_log(LOG_DEBUG, "%s: %s [%s] = %s\n", logstr, attribute_ids[i],
		datatypes[i] ? datatypes[i] : "NULL", values[i]);

        /* posix-gid attribute */
        if (strcasecmp(attribute_ids[i],
		       XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID) == 0)
        {
	    errno=0;
	    long_value=strtol(values[i], &endptr, 10);
	    if (endptr[0]!='\0' || errno!=0 || long_value<0)   {
		lcmaps_log(LOG_ERR,"%s: Cannot convert %s to a gid.\n",
			logstr, values[i]);
		return 1;
	    }
            gid = (gid_t)long_value;
            lcmaps_log(LOG_DEBUG, "%s: Adding GID :  %d\n", logstr, gid);
            if (!use_authorization_only)
                rc = addSGid(&gid, 1);
        }
        else
        {
            lcmaps_log(LOG_ERR,
		    "%s: Error: Unexpected attribute: %s [%s] = %s\n",
		    logstr, attribute_ids[i],
		    datatypes[i] ? datatypes[i] : "NULL", values[i]);
            return 1;
        }
    }

    return rc;
}


int Username_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    const char * logstr = "Username_handler";
    int rc = 0;
    int i  = 0;
    int cnt_sec_gid = 0;
    gid_t * sec_gid = NULL;
    struct passwd * pw = NULL;
    int username_attribs=0;
    xacml_decision_t decision=XACML_DECISION_Indeterminate;

    /* Get the decision from the response */
    if (xacml_response_get_xacml_decision(response, &decision)!=XACML_RESULT_SUCCESS)	{
	lcmaps_log(LOG_ERR, "%s: Error: Cannot get decision from response\n",
		logstr);
	return 1;
    }

    /* Check that the decision is Permit or Deny and matches the fullfill_on for
     * the given obligation */
    if ( ! ( (decision==XACML_DECISION_Permit &&
	      fulfill_on==XACML_EFFECT_Permit) ||
	     (decision==XACML_DECISION_Deny &&
	      fulfill_on==XACML_EFFECT_Deny) ) )
    {
	lcmaps_log(LOG_WARNING,
		"%s: Cannot fulfill obligation, "
		"fulfill on \"%s\" does not match decision \"%s\".\n",
		logstr, fulfill_on_str(fulfill_on), decision_str(decision));
	return 1;
    }
    
    lcmaps_log(LOG_DEBUG, "%s: Got obligation %s\n", logstr, obligation_id);
    setOHUsernameFired(1);

    for (i = 0; attribute_ids[i] != NULL && rc == 0; i++)
    {
	/* Check validity of value */
	if (values[i]==0 || values[i][0]=='\0') {
	    lcmaps_log(LOG_ERR,
		    "%s: Error: Attribute %s has empty value.\n",
		    logstr, attribute_ids[i]);
	    return 1;
	}

        lcmaps_log(LOG_DEBUG, "%s: %s [%s] = %s\n", logstr, attribute_ids[i],
		datatypes[i] ? datatypes[i] : "NULL", values[i]);

        /* Username attribute */
        if (strcasecmp(attribute_ids[i],
		       XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME) == 0)
        {
	    if (username_attribs!=0)	{
		lcmaps_log(LOG_ERR,
		    "%s: error: Found multiple %s attributes. "
		    "This is a protocol error.\n",
		    logstr, attribute_ids[i]);
		return 1;
	    }
	    /* Set flag */
	    username_attribs=1;

            pw = getpwnam(values[i]);
            if (!pw)
            {
                lcmaps_log(LOG_ERR,
		    "%s: Error: Couldn't find the username '%s' on the system.\n",
		    logstr, values[i]);
                return 1;
            }

            lcmaps_log (LOG_DEBUG, "%s: Found username %s = %d\n",
		    logstr, values[i], pw->pw_uid);
	    lcmaps_log(LOG_DEBUG, "%s: Adding UID :  %d\n", logstr, pw->pw_uid);
	    lcmaps_log(LOG_DEBUG, "%s: Adding GID :  %d\n", logstr, pw->pw_gid);
	    if (!use_authorization_only) {
		rc = addUid(pw->pw_uid);
		rc = addPGid(pw->pw_gid);
	    }

	    /*
	     * Retrieve secondary group id's
	     */
	    if (lcmaps_get_gidlist(values[i], &cnt_sec_gid, &sec_gid)==0)
	    {
		lcmaps_log(LOG_DEBUG, "%s: Adding %d secondary GIDs\n",
			logstr, cnt_sec_gid);
		if (!use_authorization_only)
		    rc = addSGid(sec_gid, (size_t)cnt_sec_gid);
		free(sec_gid);
            }
        }
        else
        {
            lcmaps_log(LOG_ERR,
		    "%s: Error: Unexpected attribute: %s [%s] = %s\n",
		    logstr, attribute_ids[i],
		    datatypes[i] ? datatypes[i] : "NULL", values[i]);
            return 1;
        }
    }

    return rc;
}


int Account_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    const char * logstr = "Account_handler";
    int rc = 0;
    int i  = 0;
    struct passwd * pw = NULL;
    struct group * gr = NULL;
    int username_attribs=0, primary_groupname_attribs=0; 
    xacml_decision_t decision=XACML_DECISION_Indeterminate;

    /* Get the decision from the response */
    if (xacml_response_get_xacml_decision(response, &decision)!=XACML_RESULT_SUCCESS)	{
	lcmaps_log(LOG_ERR, "%s: Error: Cannot get decision from response\n",
		logstr);
	return 1;
    }

    /* Check that the decision is Permit or Deny and matches the fullfill_on for
     * the given obligation */
    if ( ! ( (decision==XACML_DECISION_Permit &&
	      fulfill_on==XACML_EFFECT_Permit) ||
	     (decision==XACML_DECISION_Deny &&
	      fulfill_on==XACML_EFFECT_Deny) ) )
    {
	lcmaps_log(LOG_WARNING,
		"%s: Cannot fulfill obligation, "
		"fulfill on \"%s\" does not match decision \"%s\".\n",
		logstr, fulfill_on_str(fulfill_on), decision_str(decision));
	return 1;
    }
    
    lcmaps_log(LOG_DEBUG, "%s: Got obligation %s\n", logstr, obligation_id);
    setOHAccountFired(1);

    for (i = 0; attribute_ids[i] != NULL && rc == 0; i++)
    {
	/* Check validity of value */
	if (values[i]==0 || values[i][0]=='\0') {
	    lcmaps_log(LOG_ERR,
		    "%s: Error: Attribute %s has empty value.\n",
		    logstr, attribute_ids[i]);
	    return 1;
	}

        lcmaps_log(LOG_DEBUG, "%s: %s [%s] = %s\n", logstr, attribute_ids[i],
		datatypes[i] ? datatypes[i] : "NULL", values[i]);

	/* username attribute */
	if (strcasecmp(attribute_ids[i],
		       XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_USERNAME) == 0)
	{
	    if (username_attribs!=0)	{
		lcmaps_log(LOG_ERR,
			"%s: error: Found multiple %s attributes. "
			"This is a protocol error.\n",
			logstr, attribute_ids[i]);
		return 1;
	    }
	    /* Set flag */
	    username_attribs=1;
            if (values[i]==NULL || values[i][0]=='\0') {
                lcmaps_log(LOG_ERR,
			"%s: Error: Attribute %s has empty value."
			" This is a protocol error.\n",
			logstr, attribute_ids[i]);
                return 1;
            }
            pw = getpwnam(values[i]);
            if (!pw)	{
                lcmaps_log(LOG_ERR,
			"%s: Error: Couldn't find the username '%s' on the system.\n",
			logstr, values[i]);
                return 1;
	    }
            lcmaps_log (LOG_DEBUG, "%s: Found username %s = %d\n",
		    logstr, values[i], pw->pw_uid);
	    if (!use_authorization_only)
		rc = addUid(pw->pw_uid);
        } 
	/* primary groupname attribute */
	else if (strcasecmp(attribute_ids[i],
			    XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_PRIMARY_GROUPNAME) == 0)
	{
	    if (primary_groupname_attribs!=0)	{
		lcmaps_log(LOG_ERR,
			"%s: error: Found multiple %s attributes. "
			"This is a protocol error.\n",
			logstr, attribute_ids[i]);
		return 1;
	    }
	    /* Set flag */
	    primary_groupname_attribs=1;
            if (values[i]==NULL || values[i][0]=='\0') {
                lcmaps_log(LOG_ERR,
			"%s: Error: Attribute %s has empty value."
			" This is a protocol error.\n",
			logstr, attribute_ids[i]);
                return 1;
            }
            gr = getgrnam(values[i]);
            if (!gr)	{
                lcmaps_log(LOG_ERR,
			"%s: Error: Couldn't find the primary groupname '%s' on the system.\n",
			logstr, values[i]);
                return 1;
	    }
            lcmaps_log (LOG_DEBUG, "%s: Found primary groupname %s = %d\n",
		    logstr, values[i], gr->gr_gid);
	    if (!use_authorization_only)
		rc = addPGid(gr->gr_gid);
        } 
	/* secondary groupname attribute */
	else if (strcasecmp(attribute_ids[i],
			    XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_SECONDARY_GROUPNAME) == 0)
	{
            if (! (values[i] && (strlen(values[i]) > 0))) {
                lcmaps_log(LOG_ERR,
			"%s: Error: Attribute %s has empty value."
			" This is a protocol error.\n",
			logstr, attribute_ids[i]);
                return 1;
            }
            gr = getgrnam(values[i]);
            if (!gr)	{
                lcmaps_log(LOG_ERR,
			"%s: Error: Couldn't find the secondary groupname '%s' on the system.\n",
			logstr, values[i]);
                return 1;
	    }
            lcmaps_log (LOG_DEBUG, "%s: Found secondary groupname %s = %d\n",
		    logstr, values[i], gr->gr_gid);
	    if (!use_authorization_only)
		rc = addSGid(&(gr->gr_gid),1);
        }
	/* Other attribute: error */
	else
        {
            lcmaps_log(LOG_ERR,
		    "%s: Error: Unexpected attribute: %s [%s] = %s\n",
		    logstr, attribute_ids[i],
		    datatypes[i] ? datatypes[i] : "NULL", values[i]);
            return 1;
        }
    }

    return rc;
}




int AFSToken_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[])
{
    int i;
    const char * logstr = "AFSToken_handler";
    xacml_decision_t decision=XACML_DECISION_Indeterminate;

    /* Get the decision from the response */
    if (xacml_response_get_xacml_decision(response, &decision)!=XACML_RESULT_SUCCESS)	{
	lcmaps_log(LOG_ERR, "%s: Error: Cannot get decision from response\n",
		logstr);
	return 1;
    }

    /* Check that the decision is Permit or Deny and matches the fullfill_on for
     * the given obligation */
    if ( ! ( (decision==XACML_DECISION_Permit &&
	      fulfill_on==XACML_EFFECT_Permit) ||
	     (decision==XACML_DECISION_Deny &&
	      fulfill_on==XACML_EFFECT_Deny) ) )
    {
	lcmaps_log(LOG_WARNING,
		"%s: Cannot fulfill obligation, "
		"fulfill on \"%s\" does not match decision \"%s\".\n",
		logstr, fulfill_on_str(fulfill_on), decision_str(decision));
	return 1;
    }
    
    lcmaps_log(LOG_DEBUG, "%s: Got obligation %s\n", logstr, obligation_id);

    for (i = 0; attribute_ids[i] != NULL; i++)
    {
        lcmaps_log(LOG_DEBUG, "%s: %s [%s] = %s\n",
		logstr, attribute_ids[i], datatypes[i], values[i]);
    }
    return 1;
}
/*************************** E: Obligation Handlers *********************************/



/************************************************************
Function:     pep_register_obligation_handlers
Parameters:
              xacml_request_t   request
Description:
	      This function will register the obligations that the PEP
	      understands and it will register it with a handler function for
	      that obligation.
Parameters:
              xacml_request_t   request
return values:
      must be 0
************************************************************/
int pep_register_obligation_handlers (xacml_request_t request)
{

    /************ B: Register Obligation Handlers *********/

    /* Obligation: UID & GID (should feature two attributes) */
    addHandler(request, UIDGID_handler,
	       XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID);

    /* Obligation: Multiple Secondary GIDs in multiple attributes (requires:
     * UIDGID or Username) */
    addHandler(request, SecondaryGIDs_handler,
	       XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs);

    /* Obligation: AFS Token (requires: UIDGID or Username) */
    addHandler(request, AFSToken_handler,
	       XACML_ATTR_PROFILE_NS_URL_OBLIGATION_AFSToken);

    /* Obligation: Username (should match the UIGGID.uid attribute value) */
    addHandler(request, Username_handler,
               XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Username);

        /* Obligation: Account */
    addHandler(request, Account_handler,
               XACML_ATTR_PROFILE_NS_URL_OBLIGATION_Account);

#ifdef STAKEHOLDER_OSG

    /* Obligation: Path restriction (requires: UIDGID or Username) */
    addHandler(request, RootAndHomePaths_handler,
               XACML_ATTR_PROFILE_NS_URL_OBLIGATION_RootAndHomePaths);

    /* Obligation: Storage priorities (requires: UIDGID or Username) */
    addHandler(request, StorageAccessPriority_handler,
               XACML_ATTR_PROFILE_NS_URL_OBLIGATION_StorageAccessPriority);

    /* Obligation: Storage Access Permissions (requires: (UIDGID or Username)
     * and RootAndHomePaths ) */
    addHandler(request, AccessPermissions_handler,
               XACML_ATTR_PROFILE_NS_URL_OBLIGATION_AccessPermissions);
#endif /* STAKEHOLDER == "OSG" */

/*
    addHandler(request, local_user_name_handler,
            "urn:globus:local-user-name:obj");
*/
    /*** Default handler: this handler will be used for fault handling of the obligation that are not supported ***/
    addHandler(request, default_handler, NULL);
    /************ E: Register Obligation Handlers *********/

    return 0;
}



/************************************************************
Function:     pep_construct_request_subject
Parameters:
Description:
Parameters:
return values:
      must be 0
************************************************************/
int pep_construct_request_subject (xacml_request_t   request,
			    subject_attr_t    *d,
			    lcmaps_vomsdata_t *vomsdata,
			    int               cert_client_side_verified,
			    int               voms_client_side_verified
                          )
{
    const char *logstr = "pep_construct_request_subject()";
    int         i,j;
    char	*eff_subject_x509_issuer=NULL;
    char	*eff_voms_issuer=NULL, *eff_voms_issuer_ca=NULL;

    /************ B: Subject *********/

    /* Required operation */

    /* Check validity subject_x509_id */
    if (! d->subject_x509_id) {
        lcmaps_log(LOG_ERR,
		   "%s: Error: No subject DN found, this element is mandatory\n",
		   logstr);
        return 1;
    }

    lcmaps_log(LOG_DEBUG,
	    "%s: Setting XACML request subject attributes.\n", logstr);

    xacml_request_set_subject(
            request,
            d->subject_x509_id);

    /* */
    switch(cert_client_side_verified)	{
	case -1:    /* undefined */
	    eff_subject_x509_issuer="";
	    break;
	case 0:	    /* unverified */
	    eff_subject_x509_issuer=XACML_ATTR_PROFILE_NS_URL_ISSUER_NONE;
	    break;
	case 1:	    /* subject-x509-issuer */
	    eff_subject_x509_issuer=d->subject_x509_issuer;
	    break;
	default:    /* internal error */
	    lcmaps_log(LOG_ERR,
		    "%s: Error: unexpected value for cert_client_side_verified: %d\n",
		    logstr,cert_client_side_verified);
    }

    /* Non-VOMS attributes */

    /* subject-x509-id */
    lcmaps_log(LOG_DEBUG,"    %-25s = %s\n",
		"subject-x509-id", d->subject_x509_id);
    xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ID,
            XACML_DATATYPE_STRING,
            eff_subject_x509_issuer,
            d->subject_x509_id);

    if (d->subject_x509_issuer)	{
	lcmaps_log(LOG_DEBUG,"    %-25s = %s\n",
		"subject-x509-issuer", d->subject_x509_issuer);
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ISSUER,
            XACML_DATATYPE_STRING,
            d->ca_issuer_dn,
            d->subject_x509_issuer);
    }

    if (d->certificate_serial_number) {
	lcmaps_log(LOG_DEBUG,"    %-25s = %s\n",
		"certificate-serial-number", d->certificate_serial_number);
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_CERTIFICATE_SERIAL_NUMBER,
            XACML_DATATYPE_STRING,
            eff_subject_x509_issuer,
            d->certificate_serial_number);
    }
    
    if (d->ca_serial_number) {
	lcmaps_log(LOG_DEBUG,"    %-25s = %s\n",
		"ca-serial-number", d->ca_serial_number);
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_CA_SERIAL_NUMBER,
            XACML_DATATYPE_INTEGER,
            d->ca_issuer_dn,
            d->ca_serial_number);
    }

    if (d->validity_not_before[0] != '\0') {
	lcmaps_log(LOG_DEBUG,"    %-25s = %s\n",
		"validity-not-before", d->validity_not_before);
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_VALIDITY_NOT_BEFORE,
            XACML_DATATYPE_DATE_TIME,
            eff_subject_x509_issuer,
            d->validity_not_before);
    }

    if (d->validity_not_after[0] != '\0') {
	lcmaps_log(LOG_DEBUG,"    %-25s = %s\n",
		"validity-not-after", d->validity_not_after);
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
              XACML_ATTR_PROFILE_NS_URL_VALIDITY_NOT_AFTER,
            XACML_DATATYPE_DATE_TIME,
            eff_subject_x509_issuer,
            d->validity_not_after);
    }

    for (i=0; i<d->numpols; i++)   {
	lcmaps_log(LOG_DEBUG,"    ca-policy-oid [%d]         = %s\n",
		i,d->ca_policy_oid[i]); 
	xacml_request_add_subject_attribute(
	    request,
	    XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
	    XACML_ATTR_PROFILE_NS_URL_CA_POLICY_OID,
	    XACML_DATATYPE_STRING,
	    eff_subject_x509_issuer,
	    d->ca_policy_oid[i]);
    }

    if (d->cert_chain)	{
	lcmaps_log(LOG_DEBUG,"    %-25s = <PEM string of %lu bytes>\n",
		"cert-chain", (long unsigned)strlen(d->cert_chain));
        xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_CERT_CHAIN,
            XACML_DATATYPE_STRING,
            eff_subject_x509_issuer,
            d->cert_chain);
    }

    /* VOMS Attributes */
    if (vomsdata)    {
	/* Loop over all VOs */
	for (i=0; i<vomsdata->nvoms; i++)   {
	    /* Find right issuer */
	    switch(voms_client_side_verified)	{
		case -1:    /* undefined */
		    eff_voms_issuer="";
		    break;
		case 0:	    /* unverified */
		    eff_voms_issuer=XACML_ATTR_PROFILE_NS_URL_ISSUER_NONE;
		    break;
		case 1:	    /* subject-x509-issuer */
		    eff_voms_issuer=vomsdata->voms[i].voms_issuer_dn;
		    eff_voms_issuer_ca=vomsdata->voms[i].voms_issuer_ca;
		    break;
		default:    /* internal error */
		    lcmaps_log(LOG_ERR,
			    "%s: Error: unexpected value for voms_client_side_verified: %d\n",
			    logstr,voms_client_side_verified);
	    }

	    /* VO */
	    if (vomsdata->voms[i].voname)   {
		lcmaps_log(LOG_DEBUG,"    %s [%d]                    = %s\n",
			"vo",i,vomsdata->voms[i].voname);
		xacml_request_add_subject_attribute(
		    request,
		    XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
		    XACML_ATTR_PROFILE_NS_URL_VO,
		    XACML_DATATYPE_STRING,
		    eff_voms_issuer,
		    vomsdata->voms[i].voname);
	    }

	    /* voms-signing-subject */
	    if (vomsdata->voms[i].voms_issuer_dn)   {
		lcmaps_log(LOG_DEBUG,"    voms-signing-subject [%d]  = %s\n",
			i, vomsdata->voms[i].voms_issuer_dn);
		xacml_request_add_subject_attribute(
		    request,
		    XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
		    XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_SUBJECT,
		    XACML_DATATYPE_STRING,
		    eff_voms_issuer_ca,
		    vomsdata->voms[i].voms_issuer_dn);
	    }

	    /* voms-signing-issuer */
	    if (vomsdata->voms[i].voms_issuer_ca)   {
		lcmaps_log(LOG_DEBUG,"    voms-signing-issuer [%d]   = %s\n",
			i, vomsdata->voms[i].voms_issuer_ca);
		xacml_request_add_subject_attribute(
		    request,
		    XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
		    XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_ISSUER,
		    XACML_DATATYPE_STRING,
		    "",
		    vomsdata->voms[i].voms_issuer_ca);
	    }

	    /* voms-dns-port */
	    if (vomsdata->voms[i].uri)	{
		lcmaps_log(LOG_DEBUG,"    voms-dns-port [%d]         = %s\n",
			i,vomsdata->voms[i].uri);
		xacml_request_add_subject_attribute(
		    request,
		    XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
		    XACML_ATTR_PROFILE_NS_URL_VOMS_DNS_PORT,
		    XACML_DATATYPE_STRING,
		    eff_voms_issuer,
		    vomsdata->voms[i].uri);
	    }

	    /* voms-fqan */
	    for (j=0; j<vomsdata->voms[i].nfqan; j++)	{
		lcmaps_log(LOG_DEBUG,"    voms-fqan [%d][%d]          = %s\n",
			i,j, vomsdata->voms[i].fqan_unix[j].fqan ?
			    vomsdata->voms[i].fqan_unix[j].fqan : "");
		xacml_request_add_subject_attribute(
		    request,
		    XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
		    XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_FQAN,
		    XACML_DATATYPE_STRING,
		    eff_voms_issuer,
		    vomsdata->voms[i].fqan_unix[j].fqan ?
			vomsdata->voms[i].fqan_unix[j].fqan : "");
	    }

	    /* voms-primary-fqan */
	    if (i==0)	{
		lcmaps_log(LOG_DEBUG,"    voms-primary-fqan [%d]     = %s\n",
			i, vomsdata->voms[i].fqan_unix[0].fqan ?
			    vomsdata->voms[i].fqan_unix[0].fqan : "");
		xacml_request_add_subject_attribute(
		    request,
		    XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
		    XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_PRIMARY_FQAN,
		    XACML_DATATYPE_STRING,
		    eff_voms_issuer,
		    vomsdata->voms[i].fqan_unix[0].fqan ?
			vomsdata->voms[i].fqan_unix[0].fqan : "");
	    }
	} /* vo-loop */
    }

    /************ E: Subject *********/


    return 0;
}


int pep_construct_request_resource (xacml_request_t   request,
                                    const char *      resource_type,
                                    const char *      host_dns_name,
                                    const char *      resource_id,
                                    const char *      resource_x509_id,
                                    const char *      resource_x509_issuer
                                   )
{
    xacml_resource_attribute_t ra;

    /************ B: Resource *********/

    /* Node type (WN, CE, SE, RB) */
    /* Note: The Resource Type is names by the the 'Resource-id'.
     *       this was the XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE.
     *       This is to use more of the already standardized attribute namespaces.
     */
    if (resource_type)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID,
                XACML_DATATYPE_STRING,
                "",
                resource_type);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }

    /* DNS Host name */
    if (host_dns_name)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_DNS_HOST_NAME,
                XACML_DATATYPE_STRING,
                "",
                host_dns_name);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }

    /* resource-id */
    if (resource_id)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID,
                XACML_DATATYPE_STRING,
                "",
                resource_id);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }

    /* resource-x509-id */
    if (resource_x509_id)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ID,
                XACML_DATATYPE_STRING,
                "",
                resource_x509_id);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }
    /************ E: Resource *********/

    /* resource-x509-issuer */
    if (resource_x509_issuer)
    {
        xacml_resource_attribute_init(&ra);
        xacml_resource_attribute_add(
                ra,
                XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ISSUER,
                XACML_DATATYPE_STRING,
                "",
                resource_x509_issuer);
        xacml_request_add_resource_attribute(request, ra);
        xacml_resource_attribute_destroy(ra);
    }
    /************ E: Resource *********/

    return 0;
}


int pep_construct_request_action   (xacml_request_t   request,
                                    const char *      action_type
                                   )
{
    /************ B: Action *********/
    if (action_type)
    {
        xacml_request_add_action_attribute(
                request,
                XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID,
                XACML_DATATYPE_STRING,
                "",
                action_type);
    }
    /************ E: Action *********/

    return 0;
}

/* Returns whether the UIDGID OH has fired */
int getOHUIDGIDFired(void) {
        return oh_uidgid_fired;
}
/* Returns whether the SecondaryGID OH has fired */
int getOHSecondaryGIDsFired(void) {
        return oh_secondarygids_fired;
}
/* Returns whether the Username OH has fired */
int getOHUsernameFired(void) {
        return oh_username_fired;
}
/* Returns whether the Account OH has fired */
int getOHAccountFired(void) {
        return oh_account_fired;
}

/* Sets that the UIDGID OH has fired */
static void setOHUIDGIDFired(int fired) {
        oh_uidgid_fired = fired;
}
/* Sets that the SecondaryGID OH has fired */
static void setOHSecondaryGIDsFired(int fired) {
        oh_secondarygids_fired = fired;
}
/* Sets that the Username OH has fired */
static void setOHUsernameFired(int fired) {
        oh_username_fired = fired;
}
/* Sets that the Account OH has fired */
static void setOHAccountFired(int fired) {
	oh_account_fired = fired;
}


/**
 * Returns list and number of UIDs as parameters,
 * returns -1 on error, 0 on success
 */
int get_uid_list(uid_t *uids[], size_t *count)	{
    if (uids==NULL || count==NULL)
	return -1;

    *uids=uid_list;
    *count=cnt_uid;

    return 0;
}

/**
 * Returns list and number of primary GIDs as parameters,
 * returns -1 on error, 0 on success
 */
int get_pgid_list(gid_t *pgids[], size_t *count)    {
    if (pgids==NULL || count==NULL)
	return -1;

    *pgids=pgid_list;
    *count=cnt_pgid;

    return 0;
}

/**
 * Returns list and number of secondary GIDs as parameters,
 * returns -1 on error, 0 on success
 */
int get_sgid_list(gid_t *sgids[], size_t *count)    {
    if (sgids==NULL || count==NULL)
	return -1;

    *sgids=sgid_list;
    *count=cnt_sgid;

    return 0;
}

/**
 * Free and re-initialize internal lists of UIDs, pGIDs and sGIDs
 */
void clean_uidgid_lists(void)	{
    free(uid_list); uid_list=NULL;
    free(pgid_list); pgid_list=NULL;
    free(sgid_list); sgid_list=NULL;

    cnt_uid=max_cnt_uid=0;
    cnt_pgid=max_cnt_pgid=0;
    cnt_sgid=max_cnt_sgid=0;
}

/**
 * Converts decision to a string
 */
const char *decision_str(xacml_decision_t decision)	{
    switch (decision)	{
	case XACML_DECISION_Permit:
	    return "Permit";
	case XACML_DECISION_Deny:
	    return "Deny";
	case XACML_DECISION_Indeterminate:
	    return "Indeterminate";
	case XACML_DECISION_NotApplicable:
	    return "NotApplicable";
	default:
	    return "Unknown decision";
    }
}

/**
 * Converts fulfill_on to a string
 */
const char *fulfill_on_str(xacml_effect_t fulfill_on)	{
    switch (fulfill_on)	{
	case XACML_EFFECT_Permit:
	    return "Permit";
	case XACML_EFFECT_Deny:
	    return "Deny";
	default:
	    return "Unknown fulfill_on";
    }
}


/************************************************************************
 * Private functions
 ************************************************************************/

/* Adds given uid to internal list of received UIDs */
static int addUid(uid_t uid)    {
    uid_t *new_uid_list;
    size_t new_max_cnt_uid=max_cnt_uid;

    /* Increase global list when needed */
    if (cnt_uid>=max_cnt_uid)	{
	new_max_cnt_uid=max_cnt_uid+10;
	new_uid_list=(uid_t*)realloc(uid_list, new_max_cnt_uid*sizeof(uid_t));
	if (new_uid_list==NULL)	{
	    lcmaps_log(LOG_ERR,
		    "%s: out of memory trying to add uid\n", __func__);
	    return -1;
	}
	max_cnt_uid=new_max_cnt_uid;
	uid_list=new_uid_list;
    }
   
    /* Add uid */
    uid_list[cnt_uid++]=uid;

    return 0;
}

/* Adds given gid to internal list of received primary GIDs */
static int addPGid(gid_t pgid)	{
    gid_t *new_pgid_list;
    size_t new_max_cnt_pgid=max_cnt_pgid;

    /* Increase global list when needed */
    if (cnt_pgid>=max_cnt_pgid)	{
	new_max_cnt_pgid=max_cnt_pgid+10;
	new_pgid_list=(gid_t*)realloc(pgid_list, new_max_cnt_pgid*sizeof(gid_t));
	if (new_pgid_list==NULL)	{
	    lcmaps_log(LOG_ERR,
		    "%s: out of memory trying to add pgid\n", __func__);
	    return -1;
	}
	max_cnt_pgid=new_max_cnt_pgid;
	pgid_list=new_pgid_list;
    }
   
    /* Add pgid */
    pgid_list[cnt_pgid++]=pgid;

    return 0;
}

/* Adds given list of gids to internal list of received secondary GIDs */
static int addSGid(gid_t sgid[], size_t count)	{
    gid_t *new_sgid_list;
    size_t i, new_max_cnt_sgid=max_cnt_sgid;

    /* Increase global list when needed */
    if (cnt_sgid+count>max_cnt_sgid)	{
	new_max_cnt_sgid=max_cnt_sgid+count+10;
	new_sgid_list=(gid_t*)realloc(sgid_list, new_max_cnt_sgid*sizeof(gid_t));
	if (new_sgid_list==NULL)	{
	    lcmaps_log(LOG_ERR,
		    "%s: out of memory trying to add sgid\n", __func__);
	    return -1;
	}
	max_cnt_sgid=new_max_cnt_sgid;
	sgid_list=new_sgid_list;
    }
   
    /* Add sgid */
    for (i=0; i<count; i++) {
        lcmaps_log(LOG_DEBUG, "%s: Adding SGID :  %d\n", __func__, sgid[i]);
	sgid_list[cnt_sgid++]=sgid[i];
    }

    return 0;
}

/**
 * First adds the obligation handler for obligationID to the list, then add the
 * obligationID to the list of understood obligations
 */
static void addHandler(xacml_request_t request,
		  xacml_obligation_handler_t handler,
		  const char *obligationID)	{

    /* Add the obligation handler */
    xacml_request_add_obligation_handler(
            request,
            handler,
            NULL,
            obligationID);

    if (obligationID && strlen(obligationID)>0 )    {
	/* Add obligationID to list of understood obligations */
	xacml_request_add_environment_attribute(
	    request,
	    XACML_ATTR_PROFILE_NS_URL_ENVIRONMENT_ATTRIBUTES_OBLIG_SUPPORTED,
	    XACML_DATATYPE_STRING,
	    NULL,
	    obligationID);
    }
}
